print("import: 'glue'")
import glue

print("import: 'glue.gpstime'")
import glue.gpstime

print("import: 'glue.iterutils'")
import glue.iterutils

print("import: 'glue.lal'")
import glue.lal

print("import: 'glue.ligolw'")
import glue.ligolw

print("import: 'glue.ligolw.ilwd'")
import glue.ligolw.ilwd

print("import: 'glue.ligolw.tokenizer'")
import glue.ligolw.tokenizer

print("import: 'glue.pipeline'")
import glue.pipeline

print("import: 'glue.text_progress_bar'")
import glue.text_progress_bar

print("import: 'glue.utils'")
import glue.utils

print("import: 'glue.workflow'")
import glue.workflow

